/* config.h.  Generated automatically by configure.  */
/* config.h.in.  Generated automatically from configure.in by autoheader.  */

/* Define to empty if the keyword does not work.  */
/* #undef const */

/* Define if you have the ANSI C header files.  */
#define STDC_HEADERS 1

/* Define if you can safely include both <sys/time.h> and <time.h>.  */
#define TIME_WITH_SYS_TIME 1

/* Define if you have the gettimeofday function.  */
#define HAVE_GETTIMEOFDAY 1

/* Define if you have the select function.  */
#define HAVE_SELECT 1

/* Define if you have the usleep function.  */
#define HAVE_USLEEP 1

/* Define if you have the <sys/time.h> header file.  */
#define HAVE_SYS_TIME_H 1

/* Define if you have the <unistd.h> header file.  */
#define HAVE_UNISTD_H 1

/* Define if you have the ICE library (-lICE).  */
#define HAVE_LIBICE 1

/* Define if you have the SM library (-lSM).  */
#define HAVE_LIBSM 1

/* Define if you have the X11 library (-lX11).  */
#define HAVE_LIBX11 1

/* Define if you have the Xext library (-lXext).  */
#define HAVE_LIBXEXT 1

/* Define if you have the Xi library (-lXi).  */
#define HAVE_LIBXI 1

/* Define if you have the Xmu library (-lXmu).  */
#define HAVE_LIBXMU 1

/* Define if you have the Xt library (-lXt).  */
#define HAVE_LIBXT 1

/* Define if you have the m library (-lm).  */
#define HAVE_LIBM 1
