/*
 * pklib.h
 *
 * This file is part of the ttf2pk package.
 *
 * Code <-> Postscript tables
 *
 *    Loyer Frederic <loyer@ensta.fr>
 */

typedef unsigned char byte;


void
TFMopen(char *filename, int *charlist);

void
PKopen(char *filename, char *ident);

void
PKputglyph(int cc, int llx, int lly, int urx, int ury, 
           int w, int h, byte *b);

void
PKclose();


/* end */
